from pathlib import Path


def test_import():
    """makes sure we're importing the correct luxos package"""

    from luxos import api

    assert (Path(api.__file__).parent / "api.json").exists()

    ### number of API tests hardcoded here
    assert len(api.COMMANDS) == 71


def test_logon_required():
    from luxos import api

    assert api.logon_required("blah") is None
    assert api.logon_required("logoff") is True
    assert api.logon_required("logon") is False



"""user facing functions for normal use"""

from __future__ import annotations

import asyncio
import dataclasses as dc
import functools
import traceback
from typing import Any, Awaitable, Callable

import luxos.misc
from luxos.asyncops import rexec, validate  # noqa: F401

# we bring here functions from other modules
from luxos.exceptions import MinerCommandTimeoutError
from luxos.ips import ip_ranges, load_ips_from_csv  # noqa: F401
from luxos.syncops import execute_command  # noqa: F401

# + LuxosLaunchBaseResult
#    + LuxosLaunchResult
#    + LuxosLaunchError
#       + LuxosLaunchTimeoutError


@dc.dataclass
class LuxosLaunchBaseResult:
    host: str
    port: int

    @property
    def address(self):
        return f"{self.host}:{self.port}"


@dc.dataclass
class LuxosLaunchResult(LuxosLaunchBaseResult):
    data: Any = None


@dc.dataclass
class LuxosLaunchError(LuxosLaunchBaseResult):
    traceback: str | None = None
    brief: str = ""


@dc.dataclass
class LuxosLaunchTimeoutError(LuxosLaunchError, asyncio.TimeoutError):
    pass


async def launch(
    addresses: list[tuple[str, int]],
    function: Callable[[str, int], Awaitable[Any]],
    batch: int = 0,
    asobj: bool = False,
    callback: Callable | None = None,
) -> list[LuxosLaunchError | LuxosLaunchTimeoutError | Any]:
    """
    Launch an async function on a list of (host, port) miners.

    This function takes a list of (host, port) tuples (points) to
    miners, and for each "point" call `function` on it.

    Arguments:
        addresses: list of (host: str, port: int)
        function: async callable with (host: str, port: int) call signature
        batch: limit the number of concurrent calls (unlimited by default)
        asobj: if True all results will be instances subclasses
               of LuxosLaunchBaseResult

    Examples:
        This will gather the miners versions in a dict::

            async printme(host: str, port: int):
                res = await rexec(host, port, "version"))
            addresses = load_ips_from_csv("miners.csv")
            asyncio.run(launch(addresses, printme))


    """

    def wraps(fn):
        @functools.wraps(fn)
        async def _fn(host: str, port: int):
            out = None
            try:
                data = await fn(host, port)
                out = LuxosLaunchResult(host, port, data) if asobj else data
            except (asyncio.TimeoutError, MinerCommandTimeoutError) as exc:
                tback = "".join(traceback.format_exc())
                brief = repr(exc.__context__ or exc.__cause__)
                out = LuxosLaunchTimeoutError(host, port, traceback=tback, brief=brief)
            except Exception as exc:
                tback = "".join(traceback.format_exc())
                brief = repr(exc.__context__ or exc.__cause__)
                out = LuxosLaunchError(host, port, traceback=tback, brief=brief)
            return out

        return _fn

    call = wraps(function)
    if batch:
        result = []
        for subaddresses in luxos.misc.batched(addresses, batch):
            tasks = [call(*address) for address in subaddresses]
            result.extend(await asyncio.gather(*tasks, return_exceptions=True))
            if callback:
                callback(result)
        return result
    else:
        tasks = [call(*address) for address in addresses]
        return await asyncio.gather(*tasks, return_exceptions=True)



from __future__ import annotations

import argparse
import sys
import types
from typing import Callable

# The return type of add_arguments

if sys.version_info >= (3, 10):
    ArgsCallback = Callable[[argparse.Namespace], None | argparse.Namespace]
else:
    ArgsCallback = Callable


# The luxor base parser
class LuxosParserBase(argparse.ArgumentParser):
    def __init__(self, modules: list[types.ModuleType], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modules = modules
        self.callbacks: list[ArgsCallback | None] = []

    def add_argument(self, *args, **kwargs):
        try:
            if issubclass(kwargs.get("type"), ArgumentTypeBase):
                kwargs["default"] = kwargs.get("type")(kwargs.get("default"))
                kwargs["type"] = kwargs["default"]
        except TypeError:
            pass
        super().add_argument(*args, **kwargs)


class ArgumentTypeBase:
    class _NA:
        pass

    def __init__(self, default=_NA):
        self.default = default
        if default is not ArgumentTypeBase._NA:
            self.default = self._validate(default)

    def __call__(self, txt):
        self._value = None
        self._value = self._validate(txt)
        return self

    @property
    def value(self):
        return getattr(self, "_value", self.default)

    def _validate(self, value):
        try:
            return self.validate(value)
        except argparse.ArgumentTypeError as exc:
            if not hasattr(self, "_value"):
                raise RuntimeError(f"cannot use {value=} as default: {exc.args[0]}")
            raise

    def validate(self, txt):
        raise NotImplementedError("need to implement the .validate(self, txt)  method")



import argparse
import sys
from unittest import mock

from luxos.cli import v1 as cli


def test_docstring():
    @cli.cli()
    def hello():
        "this is a docstring"
        pass

    assert hello.__doc__ == "this is a docstring"


def test_callme():
    @cli.cli()
    def main(parser):
        assert isinstance(parser, argparse.ArgumentParser)

    main()

    args = ["dummy.py"]
    with mock.patch.object(sys, "argv", args):

        @cli.cli()
        def main2(args):
            assert isinstance(args, argparse.Namespace)

        main2()


def test_wrapped_cli():
    @cli.cli()
    def main(parser):
        """a simple main function

           Lorem Ipsum is simply dummy text of the printing and
         typesetting industry. Lorem Ipsum has been the industry's
        dummy text ever since the 1500s, when an unknown printer
          took a galley of type and scrambled it to make a type specimen book.
        """
        assert isinstance(parser, argparse.ArgumentParser)

    @cli.cli()
    def main2(parser):
        """another simple main function

        Lorem Ipsum is simply dummy text of the printing and
            typesetting industry. Lorem Ipsum has been the industry's standard
          dummy text ever since the 1500s, when an unknown printer
        took a galley of type and scrambled it to make a type specimen book.

        """
        assert isinstance(parser, argparse.ArgumentParser)

    assert main.attributes["doc"] == main.__doc__
    assert main2.attributes["doc"] == main2.__doc__



import argparse
import datetime

import pytest

from luxos.cli import flags


def test_type_range(resolver):
    assert flags.type_range("127.0.0.1") == [("127.0.0.1", None)]
    assert flags.type_range("127.0.0.1:1234") == [("127.0.0.1", 1234)]
    assert flags.type_range("127.0.0.1:127.0.0.3") == [
        ("127.0.0.1", None),
        ("127.0.0.2", None),
        ("127.0.0.3", None),
    ]

    assert flags.type_range("a.host") == [
        ("a.host", None),
    ]
    assert flags.type_range("a.host:111") == [
        ("a.host", 111),
    ]

    # many ways to express the same range
    fmts = [
        "127.0.0.1:1234-127.0.0.2",
        "127.0.0.1:1234:127.0.0.2",
        "127.0.0.1:1234:127.0.0.2:1234",
        "127.0.0.1:127.0.0.2:1234",
        "127.0.0.1-127.0.0.2:1234",
        "127.0.0.1:1234-127.0.0.2:1234",
    ]
    for fmt in fmts:
        assert flags.type_range(fmt) == [
            ("127.0.0.1", 1234),
            ("127.0.0.2", 1234),
        ]

    pytest.raises(argparse.ArgumentTypeError, flags.type_range, "12")
    pytest.raises(argparse.ArgumentTypeError, flags.type_range, "a.host:another.host")


def test_type_hhmm():
    assert flags.type_hhmm("12:34").default == datetime.time(12, 34)

    pytest.raises(RuntimeError, flags.type_hhmm, "12")
    pytest.raises(argparse.ArgumentTypeError, flags.type_hhmm(), "12")


def test_type_ipaddress():
    assert flags.type_ipaddress("hello").default == ("hello", None)
    assert flags.type_ipaddress("hello:123").default == ("hello", 123)

    pytest.raises(RuntimeError, flags.type_ipaddress, "12:dwedwe")
    pytest.raises(argparse.ArgumentTypeError, flags.type_ipaddress(), "12:dwedwe")



"""Script to issue commands to miner(s)

This tool is designed to issue a simple command (with parameters) to a set of miners
defined on the command line (--range/--range_start/--range-end) or in a file (yaml/csv).

It loads a `script` file and execute across all the defined miners.

Eg.
```shell

# my-script.py
from luxos import asyncops
async def main(host: str, port: int):
    res = await asyncops.rexec(host, port, "version")
    return asyncops.validate(host, port, res, "VERSION")[0]

# in the cli
$> luxos-run --range 127.0.0.1 my-script.py --json
```

"""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
import pickle
from pathlib import Path

from luxos import misc, text, utils

from ..cli import v1 as cli

log = logging.getLogger(__name__)


def add_arguments(parser: cli.LuxosParserBase) -> None:
    cli.flags.add_arguments_new_miners_ips(parser)
    cli.flags.add_arguments_rexec(parser)
    parser.add_argument("script", type=Path, help="python script to run")
    parser.add_argument(
        "-e",
        "--entry-point",
        dest="entrypoint",
        help="script entry point",
        default="main",
    )
    parser.add_argument(
        "-t", "--tear-donw", dest="teardown", help="script tear down function"
    )
    parser.add_argument("-n", "--batch", type=int, help="limit parallel executions")
    parser.add_argument(
        "--list", action="store_true", help="just display the machine to run script"
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--json", action="store_true", help="json output")
    group.add_argument("--pickle", type=Path, help="pickle output")


def process_args(args: argparse.Namespace):
    if not args.addresses:
        args.error("need a miners flag (eg. --range/--ipfile)")
    if not args.script.exists():
        args.error(f"missings script file {args.script}")


@cli.cli(add_arguments=add_arguments, process_args=process_args)
async def main(args: argparse.Namespace):
    if args.list:
        for address in args.addresses:
            print(address)
        return
    module = misc.loadmod(args.script)

    entrypoint = getattr(module, args.entrypoint, None)
    if not entrypoint:
        args.error(f"no entry point {args.entrypoint} in {args.script}")
        return

    teardown = None
    if args.teardown:
        if not hasattr(module, args.teardown):
            args.error(f"no tear down function {args.teardown} in {args.script}")
        teardown = getattr(module, args.teardown, None)

    result = {}

    def callback(result):
        log.info("processed %i / %i", len(result), len(args.addresses))

    for data in await utils.launch(
        args.addresses, entrypoint, batch=args.batch, asobj=True, callback=callback
    ):
        if isinstance(data, utils.LuxosLaunchTimeoutError):
            log.warning(
                "failed connection to %s: %s\n%s",
                data.address,
                data.brief,
                text.indent(data.traceback or "", "| "),
            )
        elif isinstance(data, utils.LuxosLaunchError):
            log.warning(
                "insternal error from %s: %s\n%s",
                data.address,
                data.brief,
                text.indent(data.traceback or "", "| "),
            )
        else:
            result[data.address] = data.data

    if args.json:
        print(json.dumps(result, indent=2))
    if args.pickle:
        args.pickle.write_bytes(pickle.dumps(result))

    if teardown:
        teardown()


def run():
    asyncio.run(main())


if __name__ == "__main__":
    run()



"""Script to issue commands to miner(s)

This tool is designed to issue a simple command (with parameters) to a set of miners
defined on the command line (--range/--range_start/--range-end) or in a file (yaml/csv).
"""

from __future__ import annotations
import asyncio
import logging
import argparse
from pathlib import Path

from ..cli import v1 as cli

log = logging.getLogger(__name__)


def add_arguments(parser: cli.LuxosParserBase) -> None:

    def add_miners_arguments(group):
        group.add_argument("--range_start", help="IP start range")
        group.add_argument("--range_end", help="IP end range")
        group.add_argument(
            "--ipfile", type=Path, help="File name to store IP addresses (csv)"
        )

        group.add_argument(
            "--range",
            action="append",
            dest="addresses",
            help="IPs range or @file",
            type=cli.flags.type_range,
        )

        group.add_argument(
            "--luxos-port",
            dest="luxos_port",
            help="miners' default port",
            type=int,
            default=4028,
        )

    group = parser.add_argument_group("Miners", "miners list or range")

    add_miners_arguments(group)
    cli.flags.add_arguments_rexec(parser)

    parser.add_argument(
        "--cmd",
        "--command",
        dest="cmd",
        required=True,
        help="Command to execute on LuxOS API",
    )
    parser.add_argument(
        "--params",
        dest="parameters",
        action="append",
        help="Parameters for LuxOS API (either str or key=value pair)",
    )
    parser.add_argument(
        "--batch",
        dest="batchsize",
        type=int,
        default=100,
        help="limit execution to batch miners at time",
    )

    group1 = parser.add_mutually_exclusive_group()
    group1.add_argument(
        "-a",
        "--all",
        dest="details",
        action="store_const",
        const="all",
        help="show full result output",
    )
    group1.add_argument(
        "-j",
        "--json",
        dest="details",
        action="store_const",
        const="json",
        help="show results in json format",
    )


def process_args(args: argparse.Namespace):

    def process_addresses():  # yes, it is pretty long
        from luxos.ips import iter_ip_ranges, load_ips_from_csv

        args.addresses = [
            address for addresses in (args.addresses or []) for address in addresses
        ]

        # old way
        if args.range_start and args.range_end:
            args.addresses.extend(
                iter_ip_ranges(f"{args.range_start}-{args.range_end}")
            )
        elif args.range_start:
            args.addresses.append((args.range_start, None))
        elif args.range_end:
            args.error("--range_end requires --range_start")

        if args.ipfile:
            if args.ipfile.exists():
                args.addresses.extend(load_ips_from_csv(args.ipfile))
            else:
                args.error(f"file not found {args.ipfile}")

        args.addresses = [
            (host, port or args.luxos_port) for host, port in args.addresses
        ]

    process_addresses()
    if not args.addresses:
        args.error("need a miners flag (eg. --range_{start|end}/--range/--ipfile)")

    lparameters: list[str] = []
    dparameters: dict[str, str] = {}

    for param in args.parameters or []:
        if "=" in param:
            key, _, value = param.partition("=")
            if lparameters:
                args.error("--params can be all strings or key=value pairs, not mixed")
            dparameters[key] = value
        else:
            if dparameters:
                args.error("--params can be all strings or key=value pairs, not mixed")
            lparameters.append(param)
    args.parameters = lparameters or dparameters


@cli.cli(add_arguments=add_arguments, process_args=process_args)
async def main(args: argparse.Namespace):
    from . import async_luxos

    await async_luxos.run(
        args.addresses,
        cmd=args.cmd,
        params=args.parameters,
        batchsize=args.batchsize,
        delay=2.0,
        details=args.details or "all",
    )


def run():
    asyncio.run(main())


if __name__ == "__main__":
    run()
